/*
 * Part of SoundPalette by hikari_no_yume.
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
//! Roland GS.
//!
//! References:
//! - Roland SC-55 Owner's Manual.
//! - Roland SC-55mkII Owner's Manual.
//! - Roland SC-7 Owner's Manual (not a GS device, only has a tiny subset).

use super::{
    param_bool, param_enum, param_other, param_range, param_signed, param_unsigned,
    AddressBlockMap, ModelInfo, ParameterAddressMap,
};

/// Roland GS.
pub const GS: ModelInfo = ModelInfo {
    model_id: &[0x42],
    name: "Roland GS",
    default_device_id: 0x10, // SC-55 and SC-7 respond to this, at least
    address_size: 3,
    address_block_map: GS_ABM,
};

const GS_ABM: AddressBlockMap = &[
    (&[0x40, 0x00], "System parameters", GS_PAM_SYSTEM),
    (
        &[0x40, 0x01],
        "Patch parameters, Patch common",
        GS_PAM_PATCH_COMMON,
    ),
    (&[0x40, 0x10], "Patch parameters, Part 10", GS_PAM_PATCH),
    (&[0x40, 0x11], "Patch parameters, Part 1", GS_PAM_PATCH),
    (&[0x40, 0x12], "Patch parameters, Part 2", GS_PAM_PATCH),
    (&[0x40, 0x13], "Patch parameters, Part 3", GS_PAM_PATCH),
    (&[0x40, 0x14], "Patch parameters, Part 4", GS_PAM_PATCH),
    (&[0x40, 0x15], "Patch parameters, Part 5", GS_PAM_PATCH),
    (&[0x40, 0x16], "Patch parameters, Part 6", GS_PAM_PATCH),
    (&[0x40, 0x17], "Patch parameters, Part 7", GS_PAM_PATCH),
    (&[0x40, 0x18], "Patch parameters, Part 8", GS_PAM_PATCH),
    (&[0x40, 0x19], "Patch parameters, Part 9", GS_PAM_PATCH),
    (&[0x40, 0x1A], "Patch parameters, Part 11", GS_PAM_PATCH),
    (&[0x40, 0x1B], "Patch parameters, Part 12", GS_PAM_PATCH),
    (&[0x40, 0x1C], "Patch parameters, Part 13", GS_PAM_PATCH),
    (&[0x40, 0x1D], "Patch parameters, Part 14", GS_PAM_PATCH),
    (&[0x40, 0x1E], "Patch parameters, Part 15", GS_PAM_PATCH),
    (&[0x40, 0x1F], "Patch parameters, Part 16", GS_PAM_PATCH),
    // These controller parameters belong to the normal patch blocks in the
    // SC-55 and SC-55mkII manuals, but the prefix system here can't handle
    // that. TODO: Change it?
    (
        &[0x40, 0x20],
        "Patch parameters, Part 10 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x21],
        "Patch parameters, Part 1 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x22],
        "Patch parameters, Part 2 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x23],
        "Patch parameters, Part 3 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x24],
        "Patch parameters, Part 4 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x25],
        "Patch parameters, Part 5 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x26],
        "Patch parameters, Part 6 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x27],
        "Patch parameters, Part 7 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x28],
        "Patch parameters, Part 8 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x29],
        "Patch parameters, Part 9 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x2A],
        "Patch parameters, Part 11 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x2B],
        "Patch parameters, Part 12 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x2C],
        "Patch parameters, Part 13 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x2D],
        "Patch parameters, Part 14 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x2E],
        "Patch parameters, Part 15 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    (
        &[0x40, 0x2F],
        "Patch parameters, Part 16 (controllers)",
        GS_PAM_PATCH_CONTROLLERS,
    ),
    // TODO: Information block. (Only mentioned in SC-55 manual, not SC-55mkII.)
    // TODO: Drum setup parameters support? These have a very annoying block
    //       layout that doesn't suit the current prefix/suffix system well.
    // TODO: More specific prefixes for the Drum setup parameters?
    (&[0x41], "Drum setup parameters, MAP1", &[]),
    (&[0x41], "Drum setup parameters, MAP2", &[]),
    // TODO: Bulk dump support? Probably for reading only. A new system would be
    //       needed to support this.
    // TODO: More specific prefixes for the Bulk dump?
    (&[0x48], "Bulk dump", &[]),
    (&[0x49], "Bulk dump (Drum setup parameters)", &[]),
];

const GS_PAM_SYSTEM: ParameterAddressMap = &[
    // TODO: Proper type/range for MASTER TUNE (needs "nibblized data" support)
    param_other(&[0x00], 0x04, "MASTER TUNE", 0x00..=0x0F),
    param_unsigned(&[0x04], 0x01, "MASTER VOLUME", 0x00..=0x7F),
    param_range(
        &[0x05],
        0x01,
        "MASTER KEY-SHIFT",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_signed(&[0x06], 0x01, "MASTER PAN", 0x01..=0x7F, 0x40),
    param_enum(
        &[0x7F],
        0x01,
        // SC-55mkII name. Called "RESET TO THE GSstandard MODE" in the SC-55
        // manual.
        "MODE SET",
        // SC-55mkII manual also mentions 0x7F as valid data, but it does not
        // explain what it would do, and "MODE SET" is marked as "(Rx. only)",
        // so it can't be for querying. Perhaps that extra byte was an error,
        // it isn't in the original SC-55 manual.
        0x00..=0x00,
        &[(&[0x00], "GS Reset")],
    ),
];

const GS_PAM_PATCH_COMMON: ParameterAddressMap = &[
    // TODO: Proper type/range for PATCH NAME (needs ASCII data support)
    param_other(&[0x00], 0x10, "PATCH NAME", 0x20..=0x7F),
    // TODO: Proper type/range for VOICE RESERVE (special)
    param_other(
        &[0x10],
        0x10,
        // SC-55mkII and SC-7 name. Called "PARTIAL RESERVE" by SC-55 manual.
        "VOICE RESERVE",
        // The max value here is the polyphony limit for the particular synth:
        // 24 for the SC-55, 28 for the SC-55mkII and SC-7. Presumably even
        // larger on later models. TODO: Special handling?
        0x00..=0x18,
    ),
    param_enum(
        &[0x30],
        0x01,
        "REVERB MACRO",
        0x00..=0x07,
        &[
            (&[0x00], "Room 1"),
            (&[0x01], "Room 2"),
            (&[0x02], "Room 3"),
            (&[0x03], "Hall 1"),
            (&[0x04], "Hall 2"),
            (&[0x05], "Plate"),
            (&[0x06], "Delay"),
            (&[0x07], "Panning Delay"),
        ],
    ),
    param_unsigned(&[0x31], 0x01, "REVERB CHARACTER", 0x00..=0x07),
    param_unsigned(&[0x32], 0x01, "REVERB PRE-LPF", 0x00..=0x07),
    param_unsigned(&[0x33], 0x01, "REVERB LEVEL", 0x00..=0x7F),
    param_unsigned(&[0x34], 0x01, "REVERB TIME", 0x00..=0x7F),
    param_unsigned(&[0x35], 0x01, "REVERB DELAY FEEDBACK", 0x00..=0x7F),
    param_unsigned(&[0x36], 0x01, "REVERB SEND LEVEL TO CHORUS", 0x00..=0x7F),
    // 37h is unoccupied!
    param_enum(
        &[0x38],
        0x01,
        "CHORUS MACRO",
        0x00..=0x07,
        &[
            (&[0x00], "Chorus 1"),
            (&[0x01], "Chorus 2"),
            (&[0x02], "Chorus 3"),
            (&[0x03], "Chorus 4"),
            (&[0x04], "Feedback Chorus"),
            (&[0x05], "Flanger"),
            (&[0x06], "Short Delay"),
            (&[0x07], "Short Delay (FB)"),
        ],
    ),
    param_unsigned(&[0x39], 0x01, "CHORUS PRE-LPF", 0x00..=0x07),
    param_unsigned(&[0x3A], 0x01, "CHORUS LEVEL", 0x00..=0x7F),
    param_unsigned(&[0x3B], 0x01, "CHORUS FEEDBACK", 0x00..=0x7F),
    param_unsigned(&[0x3C], 0x01, "CHORUS DELAY", 0x00..=0x7F),
    param_unsigned(&[0x3D], 0x01, "CHORUS RATE", 0x00..=0x7F),
    param_unsigned(&[0x3E], 0x01, "CHORUS DEPTH", 0x00..=0x7F),
    param_unsigned(&[0x3F], 0x01, "CHORUS SEND LEVEL TO REVERB", 0x00..=0x7F),
];

const GS_PAM_PATCH: ParameterAddressMap = &[
    // TODO: Proper type/range for TONE NUMBER (special)
    param_other(&[0x00], 0x02, "TONE NUMBER", 0x00..=0x7F),
    param_enum(
        &[0x02],
        0x01,
        "Rx. CHANNEL",
        0x00..=0x10,
        &[
            (&[0x00], "Channel 1"),
            (&[0x01], "Channel 2"),
            (&[0x02], "Channel 3"),
            (&[0x03], "Channel 4"),
            (&[0x04], "Channel 5"),
            (&[0x05], "Channel 6"),
            (&[0x06], "Channel 7"),
            (&[0x07], "Channel 8"),
            (&[0x08], "Channel 9"),
            (&[0x09], "Channel 10"),
            (&[0x0A], "Channel 11"),
            (&[0x0B], "Channel 12"),
            (&[0x0C], "Channel 13"),
            (&[0x0D], "Channel 14"),
            (&[0x0E], "Channel 15"),
            (&[0x0F], "Channel 16"),
            (&[0x10], "OFF"),
        ],
    ),
    param_bool(&[0x03], "Rx. PITCH BEND"),
    param_bool(&[0x04], "Rx. CH PRESSURE (CAf)"),
    param_bool(&[0x05], "Rx. PROGRAM CHANGE"),
    param_bool(&[0x06], "Rx. CONTROL CHANGE"),
    param_bool(&[0x07], "Rx. POLY PRESSURE (PAf)"),
    param_bool(&[0x08], "Rx. NOTE MESSAGE"),
    param_bool(&[0x09], "Rx. RPN"),
    param_bool(&[0x0A], "Rx. NRPN"),
    param_bool(&[0x0B], "Rx. MODULATION"),
    param_bool(&[0x0C], "Rx. VOLUME"),
    param_bool(&[0x0D], "Rx. PANPOT"),
    param_bool(&[0x0E], "Rx. EXPRESSION"),
    param_bool(&[0x0F], "Rx. HOLD1"),
    param_bool(&[0x10], "Rx. PORTAMENTO"),
    param_bool(&[0x11], "Rx. SOSTENUTO"),
    param_bool(&[0x12], "Rx. SOFT"),
    param_enum(
        &[0x13],
        0x01,
        "MONO/POLY MODE",
        0x00..=0x01,
        &[(&[0x00], "Mono"), (&[0x01], "Poly")],
    ),
    param_enum(
        &[0x14],
        0x01,
        "ASSIGN MODE",
        0x00..=0x02,
        &[
            (&[0x00], "SINGLE"),
            (&[0x01], "LIMITED - MULTI"),
            (&[0x02], "FULL - MULTI"),
        ],
    ),
    param_enum(
        &[0x15],
        0x01,
        "USE FOR RHYTHM PART",
        0x00..=0x02,
        &[(&[0x00], "OFF"), (&[0x01], "MAP1"), (&[0x02], "MAP2")],
    ),
    param_range(
        &[0x16],
        0x01,
        "PITCH KEY SHIFT",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    // TODO: Proper type/range for PITCH OFFSET FINE (needs "nibblized data"
    //       support)
    param_other(&[0x17], 0x02, "PITCH OFFSET FINE", 0x00..=0x0F),
    param_unsigned(&[0x19], 0x01, "PART LEVEL", 0x00..=0x7F),
    param_unsigned(&[0x1A], 0x01, "VELOCITY SENSE DEPTH", 0x00..=0x7F),
    param_unsigned(&[0x1B], 0x01, "VELOCITY SENSE OFFSET", 0x00..=0x7F),
    // TODO: how to accomodate special "Random" value (-64) for panning?
    param_signed(&[0x1C], 0x01, "PART PANPOT", 0x00..=0x7F, 0x40),
    // TODO: MIDI note number list for these two? (share with JS?)
    param_unsigned(&[0x1D], 0x01, "KEY RANGE LOW", 0x00..=0x7F),
    param_unsigned(&[0x1E], 0x01, "KEY RANGE HIGH", 0x00..=0x7F),
    // TODO: MIDI controller number list for these two? (share with JS?)
    param_unsigned(&[0x1F], 0x01, "CC1 CONTROLLER NUMBER", 0x00..=0x5F),
    param_unsigned(&[0x20], 0x01, "CC2 CONTROLLER NUMBER", 0x00..=0x5F),
    param_unsigned(&[0x21], 0x01, "CHORUS SEND LEVEL", 0x00..=0x7F),
    param_unsigned(&[0x22], 0x01, "REVERB SEND LEVEL", 0x00..=0x7F),
    // SC-55 manual does not mention this, but SC-55mkII does. Probably added
    // with the General MIDI support? (GM mode disables bank select receive.)
    param_bool(&[0x23], "Rx. BANK SELECT [SC-55mkII+]"),
    param_signed(
        &[0x30],
        0x01,
        "TONE MODIFY 1, Vibrato rate",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x31],
        0x01,
        "TONE MODIFY 2, Vibrato depth",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x32],
        0x01,
        "TONE MODIFY 3, TVF cutoff freq.",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x33],
        0x01,
        "TONE MODIFY 4, TVF resonance",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x34],
        0x01,
        "TONE MODIFY 5, TVF & TVA Env. attack",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x35],
        0x01,
        "TONE MODIFY 6, TVF & TVA Env. decay",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x36],
        0x01,
        "TONE MODIFY 7, TVF & TVA Env. release",
        0x0E..=0x72,
        0x40,
    ),
    param_signed(
        &[0x37],
        0x01,
        "TONE MODIFY 8, Vibrato delay",
        0x0E..=0x72,
        0x40,
    ),
    // TODO: Proper type/range for SCALE TUNING (special)
    param_other(&[0x40], 0x0C, "SCALE TUNING", 0x00..=0x0F),
];

const GS_PAM_PATCH_CONTROLLERS: ParameterAddressMap = &[
    // The "cents" unit is actually written "cent" in the SC-55 and SC-55mkII
    // owner's manuals, but plural is nicer and consistent with SC-7.
    param_range(
        &[0x00],
        0x01,
        "MOD PITCH CONTROL",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_range(
        &[0x01],
        0x01,
        "MOD TVF CUT OFF CONTROL",
        0x00..=0x7F,
        0x40,
        -9600.0..=9600.0,
        "cents",
    ),
    param_range(
        &[0x02],
        0x01,
        "MOD AMPLITUDE CONTROL",
        0x00..=0x7F,
        0x40,
        -100.0..=100.0,
        "%",
    ),
    param_range(
        &[0x03],
        0x01,
        "MOD LFO1 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x04],
        0x01,
        "MOD LFO1 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x05],
        0x01,
        "MOD LFO1 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x06],
        0x01,
        "MOD LFO1 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x07],
        0x01,
        "MOD LFO2 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x08],
        0x01,
        "MOD LFO2 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x09],
        0x01,
        "MOD LFO2 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x0A],
        0x01,
        "MOD LFO2 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x10],
        0x01,
        "BEND PITCH CONTROL",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_range(
        &[0x11],
        0x01,
        "BEND TVF CUT OFF CONTROL",
        0x00..=0x7F,
        0x40,
        -9600.0..=9600.0,
        "cents",
    ),
    param_range(
        &[0x12],
        0x01,
        "BEND AMPLITUDE CONTROL",
        0x00..=0x7F,
        0x40,
        -100.0..=100.0,
        "%",
    ),
    param_range(
        &[0x13],
        0x01,
        "BEND LFO1 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x14],
        0x01,
        "BEND LFO1 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x15],
        0x01,
        "BEND LFO1 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x16],
        0x01,
        "BEND LFO1 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x17],
        0x01,
        "BEND LFO2 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x18],
        0x01,
        "BEND LFO2 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x19],
        0x01,
        "BEND LFO2 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x1A],
        0x01,
        "BEND LFO2 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x20],
        0x01,
        "CAf PITCH CONTROL",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_range(
        &[0x21],
        0x01,
        "CAf TVF CUT OFF CONTROL",
        0x00..=0x7F,
        0x40,
        -9600.0..=9600.0,
        "cents",
    ),
    param_range(
        &[0x22],
        0x01,
        "CAf AMPLITUDE CONTROL",
        0x00..=0x7F,
        0x40,
        -100.0..=100.0,
        "%",
    ),
    param_range(
        &[0x23],
        0x01,
        "CAf LFO1 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x24],
        0x01,
        "CAf LFO1 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x25],
        0x01,
        "CAf LFO1 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x26],
        0x01,
        "CAf LFO1 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x27],
        0x01,
        "CAf LFO2 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x28],
        0x01,
        "CAf LFO2 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x29],
        0x01,
        "CAf LFO2 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x2A],
        0x01,
        "CAf LFO2 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x30],
        0x01,
        "PAf PITCH CONTROL",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_range(
        &[0x31],
        0x01,
        "PAf TVF CUT OFF CONTROL",
        0x00..=0x7F,
        0x40,
        -9600.0..=9600.0,
        "cents",
    ),
    param_range(
        &[0x32],
        0x01,
        "PAf AMPLITUDE CONTROL",
        0x00..=0x7F,
        0x40,
        -100.0..=100.0,
        "%",
    ),
    param_range(
        &[0x33],
        0x01,
        "PAf LFO1 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x34],
        0x01,
        "PAf LFO1 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x35],
        0x01,
        "PAf LFO1 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x36],
        0x01,
        "PAf LFO1 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x37],
        0x01,
        "PAf LFO2 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x38],
        0x01,
        "PAf LFO2 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x39],
        0x01,
        "PAf LFO2 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x3A],
        0x01,
        "PAf LFO2 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x40],
        0x01,
        "CC1 PITCH CONTROL",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_range(
        &[0x41],
        0x01,
        "CC1 TVF CUT OFF CONTROL",
        0x00..=0x7F,
        0x40,
        -9600.0..=9600.0,
        "cents",
    ),
    param_range(
        &[0x42],
        0x01,
        "CC1 AMPLITUDE CONTROL",
        0x00..=0x7F,
        0x40,
        -100.0..=100.0,
        "%",
    ),
    param_range(
        &[0x43],
        0x01,
        "CC1 LFO1 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x44],
        0x01,
        "CC1 LFO1 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x45],
        0x01,
        "CC1 LFO1 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x46],
        0x01,
        "CC1 LFO1 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x47],
        0x01,
        "CC1 LFO2 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x48],
        0x01,
        "CC1 LFO2 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x49],
        0x01,
        "CC1 LFO2 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x4A],
        0x01,
        "CC1 LFO2 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x50],
        0x01,
        "CC2 PITCH CONTROL",
        0x28..=0x58,
        0x40,
        -24.0..=24.0,
        "semitones",
    ),
    param_range(
        &[0x51],
        0x01,
        "CC2 TVF CUT OFF CONTROL",
        0x00..=0x7F,
        0x40,
        -9600.0..=9600.0,
        "cents",
    ),
    param_range(
        &[0x52],
        0x01,
        "CC2 AMPLITUDE CONTROL",
        0x00..=0x7F,
        0x40,
        -100.0..=100.0,
        "%",
    ),
    param_range(
        &[0x53],
        0x01,
        "CC2 LFO1 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x54],
        0x01,
        "CC2 LFO1 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x55],
        0x01,
        "CC2 LFO1 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x56],
        0x01,
        "CC2 LFO1 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
    param_range(
        &[0x57],
        0x01,
        "CC2 LFO2 RATE CONTROL",
        0x00..=0x7F,
        0x40,
        -10.0..=10.0,
        "Hz",
    ),
    param_range(
        &[0x58],
        0x01,
        "CC2 LFO2 PITCH DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=600.0,
        "cents",
    ),
    param_range(
        &[0x59],
        0x01,
        "CC2 LFO2 TVF DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=2400.0,
        "cents",
    ),
    param_range(
        &[0x5A],
        0x01,
        "CC2 LFO2 TVA DEPTH",
        0x00..=0x7F,
        0x00,
        0.0..=100.0,
        "%",
    ),
];
